use std::path::PathBuf;
use std::sync::mpsc;
use std::time::Duration;

use anyhow::Result;
use crossterm::event::{self, Event, KeyCode, KeyEvent};
use ratatui::layout::{Constraint, Layout, Rect};
use ratatui::style::{Color, Style};
use ratatui::widgets::{Block, Borders, Cell, Paragraph, Row, Table, TableState};
use ratatui::Frame;
use tui_input::{Input, InputRequest};

use crate::state::{clear_selected_project_file, write_selected_project_file};
use crate::Projects;


#[derive(Debug, Clone, PartialEq, Eq)]
enum Message {
    Exit,
    Noop,
    MoveDown,
    MoveUp,
    Confirm,
    Search,
    SearchUpdate,
    ExitSearch,
}


#[derive(Debug, Default)]
enum Mode {
    #[default]
    Normal,
    Search,
}


struct State {
    projects: Projects,
    search: Input,
    mode: Mode,
    should_exit: bool,
    project_table: TableState,
    filtered_projects: Vec<String>,
    selected_project: Option<PathBuf>,
}

impl State {
    pub fn new(projects: Projects) -> Self {
        let project_table = TableState::new().with_selected(Some(0));

        Self {
            projects,
            search: Input::default(),
            mode: Mode::default(),
            should_exit: false,
            project_table,
            filtered_projects: Vec::new(),
            selected_project: None,
        }
    }
}


pub fn run(projects: Projects) -> Result<()> {
    clear_selected_project_file()?;

    let mut term = ratatui::init();
    let mut state = State::new(projects);

    let (mut msg_tx, mut msg_rx) = mpsc::channel();

    loop {
        handle_messages(&mut state, &mut msg_rx)?;

        if state.should_exit {
            break;
        }

        term.draw(|frame| draw(&mut state, frame))?;

        if !event::poll(Duration::from_millis(10))? {
            continue;
        }

        if let Event::Key(key_event) = event::read()? {
            handle_event(&mut state, &mut msg_tx, key_event);
        }
    }

    ratatui::restore();

    if let Some(selected_project) = state.selected_project {
        write_selected_project_file(selected_project)?
    }

    Ok(())
}


fn handle_event(state: &mut State, tx: &mut mpsc::Sender<Message>, event: KeyEvent) {
    let msg = match (event.code, &state.mode) {
        (KeyCode::Char('q') | KeyCode::Esc, Mode::Normal) => Message::Exit,
        (KeyCode::Char('j') | KeyCode::Down, Mode::Normal) => Message::MoveDown,
        (KeyCode::Char('k') | KeyCode::Up, Mode::Normal) => Message::MoveUp,
        (KeyCode::Char('/'), Mode::Normal) => Message::Search,

        (KeyCode::Esc, Mode::Search) => Message::ExitSearch,
        (KeyCode::Char(c), Mode::Search) => {
            state.search.handle(InputRequest::InsertChar(c));
            Message::SearchUpdate
        }
        (KeyCode::Backspace, Mode::Search) => {
            state.search.handle(InputRequest::DeletePrevChar);
            Message::SearchUpdate
        }

        (KeyCode::Enter, _) => Message::Confirm,

        _ => Message::Noop,
    };

    tx.send(msg).unwrap();
}


fn handle_messages(state: &mut State, rx: &mut mpsc::Receiver<Message>) -> Result<()> {
    let msg = match rx.try_recv() {
        Ok(msg) => msg,
        Err(mpsc::TryRecvError::Empty) => {
            return Ok(());
        }
        Err(mpsc::TryRecvError::Disconnected) => {
            state.should_exit = true;
            return Ok(());
        }
    };

    match msg {
        Message::Exit => {
            state.should_exit = true;
            state.selected_project = None;
        }
        Message::MoveDown => {
            state.project_table.select_next();
        }
        Message::MoveUp => {
            state.project_table.select_previous();
        }
        Message::Confirm => {
            if let Some(selected) = state.project_table.selected() {
                if let Some(project_path) = state.filtered_projects.get(selected) {
                    state.should_exit = true;
                    state.selected_project = Some(PathBuf::from(project_path));
                };
            }
        }
        Message::Search => {
            state.mode = Mode::Search;
        }
        Message::ExitSearch => {
            state.search = Input::default();
            state.mode = Mode::Normal;
        }
        Message::SearchUpdate => {
            state.project_table.select_first();
        }
        _ => (),
    }

    Ok(())
}


fn draw(state: &mut State, frame: &mut Frame) {
    let block = Block::default().borders(Borders::ALL);
    frame.render_widget(&block, frame.area());

    match state.mode {
        Mode::Normal => {
            let layout = Layout::vertical([Constraint::Fill(1)]);
            let inner_area = block.inner(frame.area());
            let layout_rects = layout.split(inner_area);
            draw_list(state, frame, layout_rects[0]);
        }
        Mode::Search => {
            let layout = Layout::vertical([Constraint::Fill(1), Constraint::Max(2)]);
            let inner_area = block.inner(frame.area());
            let layout_rects = layout.split(inner_area);
            draw_list(state, frame, layout_rects[0]);
            draw_search(state, frame, layout_rects[1]);
        }
    }
}


fn draw_search(state: &mut State, frame: &mut Frame, area: Rect) {
    const PROMPT_PREFIX: &str = "Search: ";

    let search_input = &state.search;

    let scroll_offset = (0, search_input.visual_scroll(area.width as usize) as u16);

    let prompt = format!("{PROMPT_PREFIX}{}", search_input.value());
    let search = Paragraph::new(prompt).scroll(scroll_offset);
    frame.render_widget(search, area);

    let cursor_pos = (
        area.x + (search_input.visual_cursor() + PROMPT_PREFIX.len()) as u16,
        area.y,
    );
    frame.set_cursor_position(cursor_pos)
}


fn draw_list(state: &mut State, frame: &mut Frame, area: Rect) {
    state.filtered_projects = state
        .projects
        .list
        .iter()
        .filter_map(|project| {
            let path_str = project.to_str().expect("invalid path string");

            let search_value = state.search.value();

            if path_str.contains(search_value) {
                Some(path_str.to_string())
            } else {
                None
            }
        })
        .collect();

    let rows = state
        .filtered_projects
        .iter()
        .map(|path| Row::new([Cell::new(path.as_str())]));

    let widths = [Constraint::Min(20)];

    let table_highlight = Style::default().fg(Color::Blue);

    let table = Table::new(rows, widths).row_highlight_style(table_highlight);

    frame.render_stateful_widget(table, area, &mut state.project_table);
}
